-- local references to common functions
local table = table
local ipairs = ipairs
local abs = math.abs
local max = math.max
local tinsert = table.insert
local tremove = table.remove

local base = _G

module("quad")

-- quadrant names
dirs = { 'nw', 'sw', 'ne', 'se' }
rdirs = { nw='se', sw='ne', ne='sw', se='nw' }

-- tree structure
root = nil
handles = {}
deadcells = {}
mincellsize = 8

-- some stats
objects = 0
livecells = 0

-- creates new cell
local function createCell(p, r, x, y, s)
  local c
  -- are there any unused cells in the pool?
  if #deadcells == 0 then
    c = {}
  else
    c = tremove(deadcells)
    c.nw = nil
    c.ne = nil
    c.sw = nil
    c.se = nil
  end
  c.parent = p
  c.relation = r
  c.x = x
  c.y = y
  c.side = s
  -- update the livecells count
  livecells = livecells + 1
  return c
end

-- destroys existing cell
local function destroyCell(c)
  -- delete reference from parent to child
  if c.parent then
    c.parent[c.relation] = nil
  end
  -- update the livecells count
  livecells = livecells - 1
  tinsert(deadcells, c)
  -- is this the root cell?
  if c == root then
    root = nil
  end
end

-- returns number of child cells
local function getChildCount(c)
  local n = 0
  for _, v in ipairs(dirs) do
    if c[v] then
      n = n + 1
    end
  end
  return n
end

-- returns true if cell has child cells
local function hasChildren(c)
  return c.nw ~= nil or c.sw ~= nil or c.ne ~= nil or c.se ~= nil
end

-- returns true if object fits entirely inside cell
local function fitsInCell(c, x, y, s)
  local dx = abs(c.x - x)
  local dy = abs(c.y - y)
  local e = c.side/4
  return s < e and dx < e and dy < e
end

-- returns child cell direction and offset
local function getDirection(c, x, y)
  local d, ox, oy
  if x < c.x then
    if y < c.y then
      d, ox, oy = 'nw', -1, -1
    else
      d, ox, oy = 'sw', -1, 1
    end
  else
    if y < c.y then
      d, ox, oy = 'ne', 1, -1
    else
      d, ox, oy = 'se', 1, 1
    end
  end
  return d, ox, oy
end

-- return cell which fits an object of given size
-- creates the cell if necessary
local function getCell(c, x, y, s)
  -- object fits inside this cell?
  local q = c.side/2
  if s*2 > q or q < mincellsize then
    return c
  end
  -- find which sub-cell the object belongs to
  local d, ox, oy = getDirection(c, x, y)
  -- create sub-cell if necessary
  if c[d] == nil then
    local e = q/2
    ox = ox*e + c.x
    oy = oy*e + c.y
    c[d] = createCell(c, d, ox, oy, q)
  end
  -- descend deeper down the tree
  return getCell(c[d], x, y, s)
end

-- select all objects in a cell
local function selectCellAll(root, dest)
  -- insert all objects in this cell
  for i, v in ipairs(root) do
    tinsert(dest, v)
  end
  for i, v in ipairs(dirs) do
    local c = root[v]
    if c then
      selectCellAll(c, dest)
    end
  end
end

-- select objects in range
local function selectCell(root, dest, x, y, hw, hh)
  -- insert all objects in this cell
  for _, v in ipairs(root) do
    tinsert(dest, v)
  end
  -- descent down the tree
  for _, v in ipairs(dirs) do
    local c = root[v]
    if c then
      local r = c.side
      local dx = abs(c.x - x)
      local dy = abs(c.y - y)
      -- query intersect this cell?
      if r > dx - hw and r > dy - hh then
        -- query covers the cell entirely?
        if r < hw - dx and r < hh - dy then
          selectCellAll(c, dest)
        else
          selectCell(c, dest, x, y, hw, hh)
        end
      end
    end
  end
end

-- trim empty cells from bottom up
local function trimBottom(c)
  while c and #c == 0 and not hasChildren(c) do
    local p = c.parent
    destroyCell(c)
    c = p
  end
end


-- trim the top of the quadtree deleting
-- root nodes if they only have a single child
function trimTop()
  while root and #root == 0 do
    -- root has one child only?
    local children = getChildCount(root)
    if children > 1 then
      return
    end
    -- get the only child node
    local child = root.nw or root.ne or root.sw or root.se
    if child == nil then
      return
    end
    -- severe the link between child and parent
    local nroot = root[child]
    nroot.relation = 'none'
    nroot.parent = nil
    -- before we remove the old root
    -- make sure it doesn't point to its only child
    root[child] = nil
    destroyCell(root)
    -- assign new root
    root = nroot
  end
end


-- insert new object
function insert(object, x, y, hw, hh)
  local s = max(hw, hh)
  --assert(s > 0)
  -- remove object from current cell
  local c = handles[object]
  if c then
    -- object remains in its current cell?
    if fitsInCell(c, x, y, s) then
      return
    end
    remove(object)
  end
  -- add root cell
  if root == nil then
    root = createCell(nil, 'none', 0, 0, s*4)
  end
  -- expand tree if necessary
  while true do
    -- can the object fit in the root cell?
    if fitsInCell(root, x, y, s) then
      local c = getCell(root, x, y, s)
      -- insert object
      tinsert(c, object)
      handles[object] = c
      objects = objects + 1
      return
    end
    -- expand tree upwards
    local d, ox, oy = getDirection(root, x, y)
    d = rdirs[d]
    local q = root.side/2
    ox = ox*q + root.x
    oy = oy*q + root.y
    -- create new root
    local nroot = createCell(nil, 'none', ox, oy, root.side*2)
    nroot[d] = root
    root.relation = d
    root.parent = nroot
    -- assign new root
    root = nroot
  end
end

-- remove object
function remove(object)
  local c = handles[object]
  -- removing a non-existing object?
  if c == nil then
    return
  end
  objects = objects - 1
  handles[object] = nil
  -- todo: make constant time
  for i, v in ipairs(c) do
    if v == object then
      tremove(c, i)
      break
    end
  end
  trimBottom(c)
end

-- select objects in range
function select(dest, x, y, hw, hh)
  if root then
    selectCell(root, dest, x, y, hw, hh);
  end
end

-- select objects in range
function selectAABB(dest, l, t, r, b)
  -- realign aabb if necessary
  if l > r then
    l, r = r, l
  end
  if t > b then
    t, b = b, t
  end
  local x, y = (l + r)/2, (t + b)/2
  local hw = (r - l)/2
  local hh = (b - t)/2
  select(dest, x, y, hw, hh)
end

-- selects all objects
function selectAll(dest, x, y, hw, hh)
  if root then
    selectCellAll(root, dest)
  end
end